

period_schema = {
    'month': ['year', 'month'],
    'week': ['year', 'month', 'week', ],
    'day': ['date'],
    'quarter': ['year', 'quarter']
}

resample_schema = {
    'second': 'S',
    'minute': 'T',
    'hour': 'H',
    'week': 'W',
    'day': 'D',
    'month': 'M',
    'quarter': 'Q'
}



import pandas as pd


def get_intervals_dict(bins, bin_names) -> dict:
    return {
        name: pd.Interval(left, right) for name, left, right in zip(
            bin_names,
            bins[:-1],
            bins[1:]
        )
    }


def get_intervals_list(bins) -> list:
    return [pd.Interval(left, right) for left, right in zip(bins[:-1], bins[1:])]


def get_period_statistic(
        ts: pd.DataFrame,
        col,
        statistic,
        periods,
        tou=None,
) -> pd.DataFrame:
    axis_names = periods
    bins = list([getattr(ts.index, period) for period in periods])
    grouped = ts.groupby(bins)[col]\
        .agg([statistic])\
        .rename_axis(axis_names)
    ts_dt = ts.copy()
    ts_dt['dt'] = ts_dt.index
    grouped['period_start'] = ts_dt['dt'].groupby(bins).min()
    # Remove non-existent bin combos by removing blanks
    # (e.g. month 2 combined with date 2013-01-01)
    grouped.dropna(subset=[statistic], inplace=True)
    return grouped




from abc import ABC, abstractmethod
from dataclasses import dataclass
import pandas as pd
import numpy as np

from ts_tariffs.tariffs import TariffRegime


class Validator:
    @staticmethod
    def electricity_data_cols(df):
        mandatory_cols = np.array([
            'demand_energy',
            'demand_power',
            'generation_energy',
            'power_factor',
            'demand_apparent'
        ])
        not_present = list([col not in df.columns for col in mandatory_cols])
        if any(not_present):
            content = ', '.join(mandatory_cols[not_present])
            raise ValueError(f'The following columns must be present in dataframe: {content}')


@dataclass
class MeterData(ABC):
    name: str
    meter_ts: pd.DataFrame

    def set_sample_rate(self, sample_rate):
        pass


@dataclass
class ElectricityMeterData(MeterData):
    units: dict

    def __post_init__(self):
        Validator.electricity_data_cols(self.meter_ts)

    @classmethod
    def from_dataframe(cls, name, df, column_map: dict):
        units = {}
        for meter_col, data in column_map.items():
            df[meter_col] = df[data['ts']]
            units[meter_col] = data['units']
        return cls(name, df[column_map.keys()], units)

    def set_sample_rate(self, sample_rate):
        pass


@dataclass
class Site:
    name: str
    tariffs: TariffRegime
    meter_data: ElectricityMeterData
    bill_ledgers: dict[pd.DataFrame] = None
    bill: dict[float] = None

    def calculate_bill(self, detailed_bill=True):
        self.bill_ledgers = {}
        self.bill = {}
        for charge in self.tariffs.charges:
            self.bill_ledgers[charge.name] = charge.calculate_charge(
                self.meter_data.meter_ts,
                detailed_bill=detailed_bill
            )
            self.bill[charge.name] = float(self.bill_ledgers[charge.name].sum())



from abc import ABC, abstractmethod
from pydantic import BaseModel, ValidationError, validate_arguments
from enum import Enum
import pandas as pd
import numpy as np
from datetime import timedelta
from typing import (
    List,
    NamedTuple,
    Type,
    Union
)

from dataclasses import dataclass

from ts_tariffs.ts_utils import get_period_statistic
from ts_tariffs.datetime_schema import period_schema, resample_schema
from ts_tariffs.units import consumption_units


def timedelta_builder(deltas: dict):
    return sum([timedelta()])


class FrequencyValidator(BaseModel):
    minutes: int
    hours: int
    days: int
    weeks: int


class TOUValidator(BaseModel):
    time_bins: List[int]
    bin_rates: List[float]
    bin_labels: List[str]


class ConsumptionUnit(str, Enum):
    kWh = 'kWh'
    kVA = 'kVA'
    day = 'day'
    month = 'month'


class Block(NamedTuple):
    min: float
    max: float


@dataclass
class Charge(ABC):
    name: str
    charge_type: str
    consumption_unit: ConsumptionUnit
    rate_unit: str
    calculate_on: str
    adjustment_factor: Union[float, None]

    def __post_init__(self):
        if not self.adjustment_factor:
            self.adjustment_factor = 1.0

    @abstractmethod
    def calculate_charge(self, meter_ts: pd.DataFrame) -> pd.DataFrame:
        """
        :param meter_ts:
        :return:
        """
        pass


@validate_arguments
@dataclass
class SingleRateCharge(Charge):
    rate: float

    def calculate_charge(
            self,
            meter_ts: pd.DataFrame,
            detailed_bill=False
    ) -> pd.DataFrame:

        bill = meter_ts.copy()
        bill[self.name] = self.adjustment_factor * meter_ts[self.calculate_on] * self.rate
        if detailed_bill:
            bill[f'rate ({self.rate_unit})'] = self.rate
            return bill
        else:
            return bill[[self.name]]


@validate_arguments
@dataclass
class ConnectionCharge(Charge):
    rate: float
    frequency_applied: str

    def calculate_charge(
            self,
            meter_ts: pd.DataFrame,
            detailed_bill=False
    ) -> pd.DataFrame:
        bill = meter_ts.copy().resample(
            resample_schema[self.frequency_applied]
        ).sum()
        bill[self.name] = self.adjustment_factor * self.rate
        bill = pd.concat([meter_ts, bill], axis=1)
        if detailed_bill:
            bill[f'rate ({self.rate_unit})'] = self.rate
            return bill
        else:
            return bill[[self.name]]


@validate_arguments
@dataclass
class TOUCharge(Charge):
    tou: TOUValidator

    def calculate_charge(
            self,
            meter_ts: pd.DataFrame,
            detailed_bill=False
    ) -> pd.DataFrame:
        prices = np.array(self.tou.bin_rates)
        bins = np.digitize(
            meter_ts.index.hour.values,
            bins=self.tou.time_bins[1:]
        )
        bill = meter_ts.copy()

        bill[self.name] = self.adjustment_factor *\
                          prices[bins] * meter_ts[self.calculate_on].to_numpy()
        if detailed_bill:
            bill['tou'] = pd.cut(
                x=meter_ts.index.hour,
                bins=self.tou.time_bins,
                labels=self.tou.bin_labels,
                ordered=False,
                include_lowest=True
            )
            bill[f'rate ({self.rate_unit})'] = prices[bins]
            return bill
        else:
            return bill[[self.name]]


@validate_arguments
@dataclass
class DemandCharge(Charge):
    #TODO: Add handler for kWh -> kVA

    rate: float
    frequency_applied: str
    tou: Union[TOUValidator, None] = None

    # def cross_validate(self):
    #     if not any([self.rate, self.tou]):
    #         raise ValidationError(
    #             f'{self.name} schema not valid: Schema for '
    #             f'DemandChargeValidator must contain either '
    #             f'a tou or rate field'
    #         )

    def calculate_charge(
            self,
            meter_ts: pd.DataFrame,
            detailed_bill=False
    ) -> pd.DataFrame:
        bill = meter_ts.copy()
        periods = period_schema[self.frequency_applied]
        bins = list([getattr(meter_ts.index, period) for period in periods])
        if self.tou:
            time_bins = pd.cut(
                meter_ts.index.hour,
                self.tou.time_bins,
                include_lowest=True
            )
            bins.append(time_bins)

            self.rate = pd.cut(
                x=meter_ts.index.hour,
                bins=self.tou.time_bins,
                labels=self.tou.bin_rates,
                ordered=False,
                include_lowest=True
            ).astype(float)

            bill['tou'] = pd.cut(
                x=meter_ts.index.hour,
                bins=self.tou.time_bins,
                labels=self.tou.bin_labels,
                ordered=False,
                include_lowest=True
            )

        bill[f'rate ({self.rate_unit})'] = self.rate
        max_idx = meter_ts.groupby(bins)[self.calculate_on].transform(max) == meter_ts[self.calculate_on]
        bill['peaks'] = meter_ts[self.calculate_on][max_idx]
        bill[self.name] = self.adjustment_factor * bill['peaks'] * bill[f'rate ({self.rate_unit})']

        if detailed_bill:
            return bill
        else:
            return bill[[self.name]]


@validate_arguments
@dataclass
class BlockCharge(Charge):
    frequency_applied: str
    blocks: List[Block]
    bin_rates: List[float]
    bin_labels: List[str]

    def calculate_charge(
            self,
            meter_ts: pd.DataFrame,
            detailed_bill=False
    ) -> pd.DataFrame:
        periods = period_schema[self.frequency_applied]
        cumulative = get_period_statistic(
            meter_ts,
            self.calculate_on,
            'sum',
            periods,
        )
        cumulative[self.name] = 0.0
        for j, block in enumerate(self.blocks):
            rate = self.bin_rates[j]
            rate_col_name = f'block_{j + 1}_energy'
            cumulative[rate_col_name] = np.clip(
                cumulative['sum'],
                block.min,
                block.max
            ) - block.min
            cumulative[self.name] += rate * cumulative[rate_col_name]
            cumulative[f'block_{j + 1}_rate ({self.rate_unit})'] = rate
        cumulative[self.name] *= self.adjustment_factor
        cumulative.set_index('period_start', inplace=True)
        bill = pd.concat([meter_ts, cumulative], axis=1)

        if detailed_bill:
            return bill
        else:
            return bill[[self.name]]


class TariffRegime:
    def __init__(
            self,
            tariff_json: dict[str] = None,
            tariff_list: List[Type[Charge]] = None,
    ):
        if tariff_json:
            self.name = tariff_json['name']
            charges = tariff_json['charges']
            self.charges = []
            errors_dict = {}
            for charge in charges:
                try:
                    # Instantiate charge with string of type and dict of attrs
                    validated_charge = globals()[charge['charge_type']](**charge)
                    self.charges.append(validated_charge)
                except ValidationError as e:
                    errors_dict[charge['name']] = e.errors()
            print(errors_dict)
        if tariff_list:
            self.charges = tariff_list


import pathlib
from setuptools import find_packages, setup
HERE = pathlib.Path(__file__).parent
README = (HERE / 'README.md').read_text()
setup(
    name='ts_tariffs',
    version='1.1.21',
    description='Calculate bills from timeseries consumption data and different tariff structures',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/uts/tariff-module',
    author='Joe Wyndham',
    author_email='joseph.wyndham@uts.edu.au',
    license='GNU Lesser General Public License v2.1',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
    ],
    packages=find_packages(exclude=('tests',)),
    include_package_data=True,
    install_requires=[
        'boto3 >= 1.18.44',
        'matplotlib >= 3.4.3',
        'numpy >= 1.21.2',
        'pandas >= 1.3.3',
        'scipy >= 1.7.1',
        'pyarrow >= 5.0.0',
        'pydantic >= 1.8.2',
   ]
)
